package net.novagamestudios.kaffeekasse.ui.login

import androidx.compose.foundation.Image
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.IntrinsicSize
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.RowScope
import androidx.compose.foundation.layout.Spacer
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.size
import androidx.compose.foundation.layout.width
import androidx.compose.foundation.text.KeyboardActions
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.automirrored.filled.Logout
import androidx.compose.material.icons.filled.ChevronRight
import androidx.compose.material.icons.filled.Key
import androidx.compose.material.icons.filled.PhonelinkErase
import androidx.compose.material.icons.filled.ScreenLockLandscape
import androidx.compose.material.icons.filled.SmartScreen
import androidx.compose.material3.AlertDialog
import androidx.compose.material3.Button
import androidx.compose.material3.Icon
import androidx.compose.material3.IconButton
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.OutlinedTextField
import androidx.compose.material3.ProvideTextStyle
import androidx.compose.material3.Text
import androidx.compose.material3.TextButton
import androidx.compose.runtime.Composable
import androidx.compose.runtime.MutableState
import androidx.compose.runtime.collectAsState
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.focus.FocusDirection
import androidx.compose.ui.platform.LocalFocusManager
import androidx.compose.ui.res.painterResource
import androidx.compose.ui.text.input.ImeAction
import androidx.compose.ui.text.input.KeyboardType
import androidx.compose.ui.unit.dp
import cafe.adriel.voyager.core.model.ScreenModel
import cafe.adriel.voyager.core.model.screenModelScope
import cafe.adriel.voyager.navigator.Navigator
import kotlinx.coroutines.launch
import net.novagamestudios.common_utils.Logger
import net.novagamestudios.common_utils.compose.components.CircularLoadingBox
import net.novagamestudios.kaffeekasse.R
import net.novagamestudios.kaffeekasse.app
import net.novagamestudios.kaffeekasse.data.cleanName
import net.novagamestudios.kaffeekasse.model.credentials.DeviceCredentials
import net.novagamestudios.kaffeekasse.model.credentials.isValid
import net.novagamestudios.kaffeekasse.model.session.Device
import net.novagamestudios.kaffeekasse.model.session.Session
import net.novagamestudios.kaffeekasse.model.session.deviceOrNull
import net.novagamestudios.kaffeekasse.repositories.LoginRepository
import net.novagamestudios.kaffeekasse.repositories.RepositoryProvider
import net.novagamestudios.kaffeekasse.repositories.i11.PortalRepository
import net.novagamestudios.kaffeekasse.ui.AppInfoTopBarAction
import net.novagamestudios.kaffeekasse.ui.FullscreenIconButton
import net.novagamestudios.kaffeekasse.ui.navigation.LoginNavigation
import net.novagamestudios.kaffeekasse.ui.util.screenmodel.ScreenModelFactory
import net.novagamestudios.kaffeekasse.ui.util.rememberSerializableState
import net.novagamestudios.kaffeekasse.ui.util.screenmodel.collectAsStateHere


class LoginScreenModel private constructor(
    private val loginRepository: LoginRepository,
    private val portal: PortalRepository
) : ScreenModel, Logger {

    val session by portal.session.collectAsStateHere()

    val isLoading by loginRepository.isPerformingAction.collectAsStateHere()

    var showLoginDeviceDialog by mutableStateOf(false)


    fun loginDevice(deviceCredentials: DeviceCredentials) {
        screenModelScope.launch {
            loginRepository.loginDevice(deviceCredentials)
            showLoginDeviceDialog = false
        }
    }

    fun logoutDevice() {
        screenModelScope.launch {
            loginRepository.logoutDevice()
        }
    }

    companion object : ScreenModelFactory<LoginNavigation.Companion, LoginScreenModel> {
        context (RepositoryProvider)
        override fun create(screen: LoginNavigation.Companion) = LoginScreenModel(
            loginRepository = loginRepository,
            portal = portalRepository
        )
    }
}



@Composable
fun LoginAdditional(
    model: LoginScreenModel,
    navigator: Navigator
) {
    val session = model.session
    when (navigator.lastItem) {
        is LoginNavigation.FormScreen -> {
            if (session is Session.WithDevice) navigator.replace(LoginNavigation.UserSelectionScreen(session.device))
        }
        is LoginNavigation.UserSelectionScreen -> {
            if (session !is Session.WithDevice) navigator.replace(LoginNavigation.FormScreen)
        }
    }
    if (model.showLoginDeviceDialog) LoginDeviceDialog(model)
}


@Composable
private fun LoginDeviceDialog(
    model: LoginScreenModel,
    modifier: Modifier = Modifier
) {
    val inputState = rememberSerializableState {
        mutableStateOf(DeviceCredentials.Empty)
    }
    val input by inputState
    AlertDialog(
        onDismissRequest = { model.showLoginDeviceDialog = false },
        confirmButton = {
            Button(
                onClick = { model.loginDevice(input) },
                enabled = !model.isLoading && input.isValid
            ) {
                Text("Einloggen")
            }
        },
        modifier,
        dismissButton = {
            TextButton(
                onClick = { model.showLoginDeviceDialog = false },
                enabled = !model.isLoading
            ) {
                Text("Abbrechen")
            }
        },
        text = {
            CircularLoadingBox(loading = model.isLoading) {
                LoginDeviceForm(
                    inputState = inputState,
                    onDone = { model.loginDevice(input) }
                )
            }
        }
    )
}

@Composable
private fun LoginDeviceForm(
    inputState: MutableState<DeviceCredentials>,
    onDone: () -> Unit,
    modifier: Modifier = Modifier
) {
    var input by inputState
    Column(
        modifier.width(IntrinsicSize.Min),
        verticalArrangement = Arrangement.spacedBy(8.dp),
        horizontalAlignment = Alignment.CenterHorizontally
    ) {
        val focusManager = LocalFocusManager.current
        OutlinedTextField(
            value = input.deviceId,
            onValueChange = { input = input.copy(deviceId = it.uppercase()) },
            label = { Text("Geräte-Id") },
            placeholder = { Text("XX:XX:XX:XX:XX:XX") },
            leadingIcon = { Icon(Icons.Default.SmartScreen, "Geräte-Id") },
            keyboardOptions = KeyboardOptions(
                keyboardType = KeyboardType.Text,
                imeAction = ImeAction.Next
            ),
            keyboardActions = KeyboardActions(onNext = { focusManager.moveFocus(FocusDirection.Next) }),
            singleLine = true
        )
        OutlinedTextField(
            value = input.apiKey,
            onValueChange = { input = input.copy(apiKey = it) },
            label = { Text("API-Schlüssel") },
            leadingIcon = { Icon(Icons.Default.Key, "API-Schlüssel") },
            keyboardOptions = KeyboardOptions(
                keyboardType = KeyboardType.Password,
                imeAction = ImeAction.Done
            ),
            keyboardActions = KeyboardActions(onDone = { onDone() }),
            singleLine = true
        )
    }
}


@Composable
fun LoginTopBarTitle(
    model: LoginScreenModel
) {
    model.session.deviceOrNull?.let {
        Row(
            verticalAlignment = Alignment.CenterVertically
        ) {
            Image(
                painterResource(R.drawable.logo_edited),
                "Kaffeekasse",
                Modifier.size(32.dp)
            )
            DeviceInfo(it, Modifier.padding(16.dp))
        }
    }
}

context (RowScope)
@Composable
fun LoginTopBarActions(
    model: LoginScreenModel
) {
    val session = model.session
    if (session !is Session.WithRealUser) {
        when (session) {
            is Session.WithDevice -> {
                var confirmLogout by remember { mutableStateOf(false) }
                IconButton(
                    onClick = { confirmLogout = true },
                    enabled = !model.isLoading
                ) {
                    Icon(Icons.Default.PhonelinkErase, "Gerät ausloggen")
                }
                if (confirmLogout) ConfirmLogoutDeviceDialog(
                    onDismiss = { confirmLogout = false },
                    onConfirm = { model.logoutDevice() }
                )
            }
            else -> {
                IconButton(
                    onClick = { model.showLoginDeviceDialog = true },
                    enabled = !model.isLoading
                ) {
                    Icon(Icons.Default.ScreenLockLandscape, "Gerät einloggen")
                }
            }
        }
    }
    AppInfoTopBarAction()
    FullscreenIconButton()
}




@Composable
private fun DeviceInfo(
    device: Device,
    modifier: Modifier = Modifier
) = Row(
    modifier,
    verticalAlignment = Alignment.CenterVertically
) {
    Text(device.name ?: "Unbekanntes Gerät")
    device.knownItemGroup?.let {
        ProvideTextStyle(MaterialTheme.typography.titleMedium) {
            Spacer(Modifier.width(16.dp))
            Icon(Icons.Default.ChevronRight, null)
            Text(
                it.cleanName,
                style = MaterialTheme.typography.titleMedium
            )
        }
    }
}

@Composable
private fun ConfirmLogoutDeviceDialog(
    onDismiss: () -> Unit,
    onConfirm: () -> Unit,
    modifier: Modifier = Modifier
) = AlertDialog(
    onDismissRequest = onDismiss,
    confirmButton = {
        Button(
            onClick = onConfirm
        ) {
            Text("Ausloggen")
        }
    },
    modifier,
    dismissButton = {
        TextButton(
            onClick = onDismiss
        ) {
            Text("Abbrechen")
        }
    },
    text = {
        Text("Möchtest du das Gerät wirklich ausloggen?")
    }
)

@Composable
fun LogoutTopBarAction(session: Session) {
    val app = app()
    val loginRepository = app.loginRepository
    val isLoading by loginRepository.isPerformingAction.collectAsState()
    CircularLoadingBox(loading = isLoading) {
        IconButton(
            onClick = { app.launch { loginRepository.logoutUser() } },
            enabled = session !is Session.Empty
        ) {
            Icon(
                Icons.AutoMirrored.Filled.Logout,
                contentDescription = "Ausloggen",
                tint = MaterialTheme.colorScheme.primary
            )
        }
    }
}

