package net.novagamestudios.kaffeekasse.ui.navigation

import androidx.compose.foundation.background
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.material3.CircularProgressIndicator
import androidx.compose.material3.MaterialTheme
import androidx.compose.runtime.Composable
import androidx.compose.runtime.LaunchedEffect
import androidx.compose.runtime.remember
import androidx.compose.ui.Modifier
import cafe.adriel.voyager.core.screen.Screen
import cafe.adriel.voyager.navigator.tab.Tab
import cafe.adriel.voyager.navigator.tab.TabNavigator
import cafe.adriel.voyager.navigator.tab.TabOptions
import net.novagamestudios.common_utils.compose.components.BoxCenter
import net.novagamestudios.kaffeekasse.model.session.Device
import net.novagamestudios.kaffeekasse.model.session.Session
import net.novagamestudios.kaffeekasse.model.session.deviceOrNull
import net.novagamestudios.kaffeekasse.ui.AppModulesScreenModel
import net.novagamestudios.kaffeekasse.ui.getValue
import net.novagamestudios.kaffeekasse.ui.hiwi_tracker.HiwiTrackerModuleScreenModel
import net.novagamestudios.kaffeekasse.ui.hiwi_tracker.HiwiTrackerTopBarActions
import net.novagamestudios.kaffeekasse.ui.hiwi_tracker.HiwiTrackerTopBarTitle
import net.novagamestudios.kaffeekasse.ui.hiwi_tracker.Overview
import net.novagamestudios.kaffeekasse.ui.hiwi_tracker.OverviewScreenModel
import net.novagamestudios.kaffeekasse.ui.kaffeekasse.Account
import net.novagamestudios.kaffeekasse.ui.kaffeekasse.AccountScreenModel
import net.novagamestudios.kaffeekasse.ui.kaffeekasse.DynamicManualBill
import net.novagamestudios.kaffeekasse.ui.kaffeekasse.KaffeekasseModuleScreenModel
import net.novagamestudios.kaffeekasse.ui.kaffeekasse.KaffeekasseTopBarActions
import net.novagamestudios.kaffeekasse.ui.kaffeekasse.KaffeekasseTopBarNavigation
import net.novagamestudios.kaffeekasse.ui.kaffeekasse.KaffeekasseTopBarTitle
import net.novagamestudios.kaffeekasse.ui.kaffeekasse.ManualBillScreenModel
import net.novagamestudios.kaffeekasse.ui.kaffeekasse.Transactions
import net.novagamestudios.kaffeekasse.ui.kaffeekasse.TransactionsScreenModel
import net.novagamestudios.kaffeekasse.ui.kaffeekasse.backNavigationHandler
import net.novagamestudios.kaffeekasse.ui.login.LoginAdditional
import net.novagamestudios.kaffeekasse.ui.login.LoginForm
import net.novagamestudios.kaffeekasse.ui.login.LoginFormScreenModel
import net.novagamestudios.kaffeekasse.ui.login.LoginScreenModel
import net.novagamestudios.kaffeekasse.ui.login.LoginTopBarActions
import net.novagamestudios.kaffeekasse.ui.login.LoginTopBarTitle
import net.novagamestudios.kaffeekasse.ui.login.UserSelection
import net.novagamestudios.kaffeekasse.ui.login.UserSelectionScreenModel
import net.novagamestudios.kaffeekasse.ui.util.screenmodel.ScreenModelProvider
import net.novagamestudios.kaffeekasse.ui.util.screenmodel.ScreenProvidingModel







sealed interface LoginNavigation {
    companion object : LoginNavigation, ScreenProvidingModel<LoginScreenModel> {
        @get:Composable override val model by LoginScreenModel
        private val initialScreen: Screen @Composable get() = model
            .session
            .deviceOrNull
            ?.let { UserSelectionScreen(it) }
            ?: FormScreen
        @Composable override fun Content() = AppScaffoldNavigator(
            key = "login-navigator",
            initialRoute = listOf(initialScreen),
            title = { LoginTopBarTitle(model) },
            actions = { LoginTopBarActions(model) },
            topAppBarScrollBehavior = null,
            content = {
                if (model.isLoading) {
                    BoxCenter(Modifier.fillMaxSize()) {
                        CircularProgressIndicator()
                    }
                } else {
                    AppScaffoldNavigatorDefaultContent("login-navigator-content")
                }
                LoginAdditional(model, this)
            }
        )
    }

    data object FormScreen : LoginNavigation, ScreenProvidingModel<LoginFormScreenModel> {
        @get:Composable override val model by LoginFormScreenModel
        @Composable override fun Content() = LoginForm(model)
    }

    data class UserSelectionScreen(
        val device: Device
    ) : LoginNavigation, ScreenProvidingModel<UserSelectionScreenModel> {
        @get:Composable override val model by UserSelectionScreenModel
        @Composable override fun Content() = UserSelection(model)
    }
}




sealed interface ModuleTab : Tab {
    val session: Session.WithRealUser
}


data class AppModulesScreen(
    val session: Session.WithRealUser
) : ScreenProvidingModel<AppModulesScreenModel> {
    @get:Composable override val model by AppModulesScreenModel
    @Composable override fun Content() {
        val model = model
        TabNavigator(
            remember { model.initialModuleTab() },
            disposeNestedNavigators = true,
            key = "app-modules-tabs",
        ) { tabNavigator ->
            AppTabTransition(
                tabNavigator = tabNavigator,
                Modifier
                    .background(MaterialTheme.colorScheme.background)
                    .fillMaxSize(),
                key = "app-modules"
            )
            val currentTab = tabNavigator.current
            LaunchedEffect(currentTab) {
                if (currentTab is ModuleTab) model.onNavigateModuleTab(currentTab)
            }
        }
    }
}


sealed interface KaffeekasseNavigation {

    data class Tab(
        override val session: Session.WithRealUser
    ) : ModuleTab, KaffeekasseNavigation, ScreenModelProvider<KaffeekasseModuleScreenModel> {
        @get:Composable override val model by KaffeekasseModuleScreenModel
        @Composable override fun Content() = AppScaffoldNavigator(
            key = "kaffeekasse-navigator",
            initialRoute = listOf(ManualBillScreen(session)),
            title = { KaffeekasseTopBarTitle(model, it) },
            navigationIcon = { KaffeekasseTopBarNavigation(model, it) },
            actions = { KaffeekasseTopBarActions(it) },
            backNavigationHandlerProvider = { model.backNavigationHandler(it) }
        )
        override val options @Composable get() = remember { TabOptions(1u, "Kaffeekasse") }
    }

    data class ManualBillScreen(
        val session: Session.WithRealUser
    ) : KaffeekasseNavigation, ScreenProvidingModel<ManualBillScreenModel> {
        @get:Composable override val model by ManualBillScreenModel
        @Composable override fun Content() = DynamicManualBill(model)
    }

    data class AccountScreen(
        val session: Session.WithRealUser
    ) : KaffeekasseNavigation, ScreenProvidingModel<AccountScreenModel> {
        @get:Composable override val model by AccountScreenModel
        @Composable override fun Content() = Account(model)
    }

    data class TransactionsScreen(
        val session: Session.WithRealUser
    ) : KaffeekasseNavigation, ScreenProvidingModel<TransactionsScreenModel> {
        @get:Composable override val model by TransactionsScreenModel
        @Composable override fun Content() = Transactions(model)
    }
}


sealed interface HiwiTrackerNavigation {

    data class Tab(
        override val session: Session.WithRealUser
    ) : ModuleTab, HiwiTrackerNavigation, ScreenModelProvider<HiwiTrackerModuleScreenModel> {
        @get:Composable override val model by HiwiTrackerModuleScreenModel
        @Composable override fun Content() = AppScaffoldNavigator(
            key = "hiwitracker-navigator",
            initialRoute = listOf(OverviewScreen(session)),
            title = { HiwiTrackerTopBarTitle(it) },
            actions = { HiwiTrackerTopBarActions(model) }
        )
        override val options @Composable get() = remember { TabOptions(2u, "Hiwi Tracker") }
    }

    data class OverviewScreen(
        val session: Session.WithRealUser
    ) : HiwiTrackerNavigation, ScreenProvidingModel<OverviewScreenModel> {
        @get:Composable override val model by OverviewScreenModel
        @Composable override fun Content() = Overview(model)
    }
}
